(*
LWSImport.pas - Enables LoGikSim to load .LWS files from LoCad for Windows
Copyright (C) 2006  Stefan Hacker <stefan.hacker(AT)web.de>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version. This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc., 51
Franklin St, Fifth Floor, Boston, MA 02110, USA
*)
unit LWSImport;

interface
uses ObjectList, Classes, Types, SysUtils, Graphics; //Get the ObjectList unit

(*
  Constant non-global stuff like IDs and so on
*)
const LWS_OR          = $6F;   //Simple or component
      LWS_XOR         = $65;   //Simple xor component
      LWS_HA          = $68;   //Half-adder
      LWS_AND         = $75;   //Simple and component
      LWS_FF          = $66;   //Flip-Flop
      LWS_VA          = $76;   //Full-adder
      LWS_SHIFT       = $73;   //Shift register
      LWS_DEC         = $64;   //Decoder
      LWS_DEC_T       = $44;   //clocked Decoder
      LWS_MUL         = $6D;   //Multiplexer
      LWS_RAM         = $72;   //Ram
      LWS_INV         = $6E;   //Inverter
      LWS_PAR6        = $70;   //Parallelport 6
      LWS_PAR8        = $50;   //Parallelport 8
      LWS_TSG         = $54;   //Tri-State-Gate
      LWS_NUM6        = $7A;   //Numeric display 6
      LWS_NUM8        = $5A;   //Numeric display 8
      LWS_BIN         = $62;   //Binary counter
      LWS_SWITCH      = $13;   //A switch
      LWS_LED         = $6C;   //A led (big and small)
      LWS_CLOCK       = $74;   //Clock generator
      LWS_DELAY       = $16;   //Signal delay

      LWS_NOTHING          = $F3;   //A clear grid point
      LWS_OBJECT           = $F4;   //A grid which is covered by an object
      LWS_CONNECTOR_HOR    = $F5;   //A Horizontal connector
      LWS_CONNECTOR_VER    = $F6;   //A Vertical connector
      LWS_TEXT             = $F9;   //Starting point of text(also ic identifier)
      LWS_CONNP_TOP        = $FB;   //Point-connector connectable from top
      LWS_CONNP_BOTTOM     = $FC;   //Point-connector connectable from bottom
      LWS_VALUE            = $FD;   //Changable value (klickable) (p.e. shifting reg.)
      LWS_LED_BODY         = $00;   //Up to now only discovered for LED bodys

      //Grid resolution
      LWS_COLUMNS     = $41;  //Std. number of rows in a locad file
      LWS_ROWS        = $2A;  //Std. number of columns in a locad file

      //Rotation states
      LWS_ROT_LEFT    = $00;  //Input @ left
      LWS_ROT_BOTTOM  = $01;  //        bottom
      LWS_ROT_RIGHT   = $02;  //        right
      LWS_ROT_TOP     = $03;  //        top

      //MODIFIERS
      ////////////
      //A single cable multiplied by 2 is recognised as on
      LWS_ON          = $02;
      //If this value is added the junction is seen as connected
      LWS_CONNECTED   = $A2;
      //If this value is added the junction is seen as coupled
      LWS_COUPLED     = $51;

      //Horizontal cable
      LWS_CABLE_HOR_L = $03;    //Left point
      LWS_CABLE_HOR_R = $1B;   //Right point
      LWS_CABLE_HOR   = LWS_CABLE_HOR_L + LWS_CABLE_HOR_R; //0x1E Horizontal cable

      //Vertical cable
      LWS_CABLE_VER_T = $09;    //Top point
      LWS_CABLE_VER_B = $01;    //Bottom point
      LWS_CABLE_VER   = LWS_CABLE_VER_T + LWS_CABLE_VER_B; //0x0A Vertical cable
      //Edges
      LWS_EDGE_L_T    = LWS_CABLE_HOR_L + LWS_CABLE_VER_T;//0x0C Cable edge left, top
      LWS_EDGE_L_B    = LWS_CABLE_HOR_L + LWS_CABLE_VER_B;//0x04 left, bottom
      LWS_EDGE_R_T    = LWS_CABLE_HOR_R + LWS_CABLE_VER_T;//0x24 right, top
      LWS_EDGE_R_B    = LWS_CABLE_HOR_R + LWS_CABLE_VER_B;//0x1C right, bottom

      //Cross sections

      LWS_CABLE_CROSS       = LWS_CABLE_HOR + LWS_CABLE_VER; //0x28 Just crossing cables
      LWS_CABLE_CROSS_CONN  = LWS_CABLE_CROSS + LWS_CONNECTED; //0xCA Connected cables

      LWS_P_NOTHING     = $00;    //Nothing
      LWS_P_TOP         = $02;
      LWS_P_BOTTOM      = $04;
      LWS_P_LEFT        = $08;
      LWS_P_RIGHT       = $10;
      LWS_P_CONNECTED   = $20;  //horizontal and vertical connected
      LWS_P_COUPLED     = $40;  //horizontal and vertical coupled

      //Colors for the LEDs
      LWS_COLOR_GREEN   = $01;
      LWS_COLOR_RED     = $02;
      LWS_COLOR_YELLOW  = $03;
(*
type LWS_P
  Datatype for flags
*)
type LWS_P = integer;
(*
type LS_LWS_Head
  This type is only for internal usage in the LS_LWSLoader class.
  It will take the full head (plus garbage) of a .LWS File containing
  not only the Grid info but also the whole block until the component
  descriptions.
*)
  LS_LWS_Head = record
  //bFirst, bSecond - Unknown up to now
  bFirst,bSecond:byte;
  //wFileID - Supposed to be a working id of the File
  wFileID:Word; //Base head

  //abMesg - Representation of the mesh with 65x42 = 2730 bytes info
  abMesh:array[1..LWS_COLUMNS,1..LWS_ROWS] of byte;
  //bColumnCount    - Number of columns in the mesh (static@ 0x2A)
  //bComponentCount - Number of components attached to the file
  bColumnCount,bComponentCount:byte;

  //Unknown stuff (to analyze)
  awUnknown:array[1..134] of byte;
end;

(*
type LS_LWS_Component
  This type is for internal usage in the LS_LWSLoader class. It will
  take the 88 Bytes info with which every single component in the .LWS
  file is described
*)
    LS_LWS_Component = record
  //bComponentType  - Defines the type of the element
  //bSubType        - Subtype of the element (norm. 0x01). Only seen with LEDs
  //                  up to now which are 0x02 when beeing big.
  bComponentType,bSubType : byte;

  bUnknown1 : byte;
  bRotation  : byte;

  //Up to know quite unknown, led position is @ 1
  rCoordinates : array [1..3] of record
    //wGridX     - X-Coordinate
    //wGridY     - Y-Coordinate
    wGridX,wGridY:Word;
    end;

  rPinIn : array[1..9] of record  // 9*6 Bytes = 54
    //wGridX      - X-Coordinate of the pin in the grid
    //wGridY      - Y-Coordinate of the pin in the grid
    //bNegated    - Is the pin negated (boolean)
    //bState      - Is the pin on (boolean)
    wGridX,wGridY:Word;
    bNegated, bState:boolean;
    end;

  wUnknown3 : Word;

  rPinOut : array[1..8] of record // 8*8 Bytes = 64
    //wGridX      - X-Coordinate of the pin in the grid
    //wGridY      - Y-Coordinate of the pin in the grid
    //bNegated    - Is the pin negated (boolean)
    //bState      - Is the pin on (boolean)
    wGridX,wGridy:Word;
    bNegated, bState:boolean;

    //bID - Eventually the ID of the output pin
    bID,bUnknown4:byte;
    end;

end;

(*
type LS_LWS_Tail
  This type is for internal usage in the LS_LWSLoader class. It will take
  the tail of the file following the body of each .LWS file.
*)
    LS_LWS_Tail = record
    //Number of attached strings
    bStrings : byte;
    //bSep1,bSep2 - Eventually a mark to recognise the tail (0x04 stock)
    bSep1,bSep2:byte;
    //sStyle - Cleartext definition of the style, maximal 26 chars + 1 byte size
    sText:string[26];
end;

(*
type LS_LWS_Text
  This type is for internal usage in the LS_LWSLoader class. It will take
  the Text element which are following the Tail of the .LWS File.
*)
    LS_LWS_Text = record
      //bGridX, bGrixY    - Position of the text in the grid
      //bSize             - Size of the Text
      bGridX, bGridY, bSize : byte;
      //sText             - Saved text, maximal 26 chars + 1 byte size
      sText:string[26];
end;

(*
type LS_Line
  This type is for internal isage in the LS_LWSLoader class. It represents
  a simple cable line.
*)
    LS_Line = record
      //pointStart  - Coordinates of the "start" point of the line
      //pointEnd    - Coordinates of the "end" point of the line
      pointStart,pointEnd:TPoint;
    end;

(*
class LS_LWSLoader
  This class represents the core of the LWSImport module. It enables you to
  load and convert existing .LWS files to the .sim format readable by the
  rest of LogikSim.
*)
    LS_LWSLoader = class(TObject)
  private
    mTElementBox:TElementBox; //Our TElementBox reference
    bOwnElementBox:boolean; //Is mTElementBox local?

    //Stuff to save .LWS data
    lwsHead:^LS_LWS_Head;   //Takes the head
    lwsComponent:TList; //Takes all components
    lwsText:TList;  //Takes all the text elements
    lwsTail:^LS_LWS_Tail;  //Takes the tail

  protected
    //Checks if pointP is in the line list and is Horizontal
    function IsInLine(pointP:TPoint;pList:TList;bHorizontal:boolean):boolean;
    //This function is supposed to gather important information of the grid
    function PropertyOf(bVal:byte):LWS_P;
    //This functions provides the real loading and is called by the public
    function LoadIn(sPathIn:string):boolean;
  public
    {
    Constructors  / Destructors
    }

    //Std constructor without parameters
    constructor Create; overload;
    //Like Load(sPathIn)
    constructor Create(sPathIn:string); overload;
    //Like Load(sPath) && SaveConverted(sPathOut)
    constructor Create(sPathIn, sPathOut:string); overload;
    //Like Load(sPath, pElementBox)
    constructor Create(sPathIn:string; pElementBox:TElementBox); overload;
    destructor Destroy; override;

    {
    Load & Save
    }

    //Loads from a file sPathIn into a virtual TElementBox
    function Load(sPathIn:string):boolean; overload;
    //Loads from a file sPathIn into an existing TElementBox pElementBox
    function Load(sPathIn:string; pElementBox:TElementBox):boolean; overload;

    //Saves the loaded .LWS file as a .sim file named sPathOut
    function SaveConverted(sPathOut:string):boolean;

    //Releases TElementBox object and resets the class
    procedure Release;

    {
    Manipulation
    }

end;



implementation
uses Haupt, RakBinaryStreamData;     //Get the Types/Classes/SysUtils unit

  {
  Constructors / Destructors
  }

  //Std constructor without parameters
  constructor LS_LWSLoader.Create;
  begin
    inherited create;
    //Set default values
    self.mTElementBox   := Nil;
    self.bOwnElementBox := False;

    self.lwsHead        := Nil;
    self.lwsComponent   := Nil;
    self.lwsText        := Nil;
    self.lwsTail        := Nil;

  end;
  //Like Load(sPathIn)
  constructor LS_LWSLoader.Create(sPathIn:string);
  begin
    self.create;
    self.Load(sPathIn);
  end;

  //Like Load(sPath) && SaveConverted(sPathOut)
  constructor LS_LWSLoader.Create(sPathIn:string; sPathOut: string);
  begin
    self.create;
    if self.Load(sPathIn) then self.SaveConverted(sPathOut);
  end;

  //Like Load(sPath, pElementBox)
  constructor LS_LWSLoader.Create(sPathIn:string; pElementBox:TElementBox);
  begin
    self.create;
    self.Load(sPathIn, pElementBox);
  end;

  destructor LS_LWSLoader.Destroy;
  begin
    self.Release;
    inherited destroy;
  end;
  {
  Load & Save
  }

  //Loads from a file sPathIn into a virtual TElementBox
  function LS_LWSLoader.Load(sPathIn:string):boolean;
  begin
    //Check if we have already loaded a file
    if self.mTElementBox <> Nil then
    begin
      Result := False;     //If not return false
      exit;
    end;
    //Be sure to be completely "fresh"
    self.Release;
    
    //Alright, create the instance to load into
    self.mTElementBox   := TElementBox.create(Nil, Nil, point(0,0));
    self.bOwnElementBox := True;
    
    //Alright, load it
    Result:=self.LoadIn(sPathIn);
  end;

  //Loads from a file sPathIn into an existing TElementBox pElementBox
  function LS_LWSLoader.Load(sPathIn:string; pElementBox:TElementBox):boolean;
  begin
    //Check if we have already loaded a file or given Box is Nil
    if (self.mTElementBox <> Nil) and (pElementBox <> Nil) then
    begin
      Result := False;     //If not return false
      exit;
    end;
    //Be sure to be completely "fresh"
    self.Release;

    self.mTElementBox := pElementBox;
    self.bOwnElementBox := False;

    Result:=self.LoadIn(sPathIn);
 end;
  function LS_LWSLoader.IsInLine(pointP:TPoint;pList:TList;bHorizontal:boolean):boolean;
  var pLine:^LS_Line;
  begin
    Result:=False;
    for pLine in pList do
      begin
        if bHorizontal then
          begin
            if (pLine.pointStart.Y = pLine.pointEnd.Y) and //Is line hor
                (pLine.pointStart.Y = pointP.Y) and  //Is colission possible
                (pLine.pointStart.X <= pointP.X) and
                (pointP.X <= pLine.pointEnd.X) then
                begin
                  Result:=True;
                  break;
                end;
          end
        else
          begin
            if (pLine.pointStart.X = pLine.pointEnd.X) and //Is line ver
                (pLine.pointStart.X = pointP.X) and  //Is colission possible
                (pLine.pointStart.Y <= pointP.Y) and
                (pointP.Y <= pLine.pointEnd.Y) then
                begin
                  Result:=True;
                  break;
                end;
          end;
      end;
  end;

  function LS_LWSLoader.PropertyOf(bVal:byte):LWS_P;
  (* This function converts the locad Cable routing description to some
  easier to interpret flagged integer.  *)
  var ret:LWS_P;
  begin
  //This is the worst hacking i've ever done but I can't think of another way
  //to handle something like that^^If someone does know a way, PLEASE tell me!

  (*
      LWS_NOTHING          = $F3;   //A clear grid point
      LWS_OBJECT           = $F4;   //A grid which is covered by an object
      LWS_CONNECTOR_HOR    = $F5;   //A Horizontal connector
      LWS_CONNECTOR_VER    = $F6;   //A Vertical connector
      LWS_TEXT             = $F9;   //Starting point of text(also ic identifier)
      LWS_CONNP_TOP        = $FB;   //Point-connector connectable from top
      LWS_CONNP_BOTTOM     = $FC;   //Point-connector connectable from bottom
      LWS_VALUE            = $FD;   //Changable value (klickable) (p.e. shifting reg.)
      LWS_LED_BODY         = $00;   //Since now only discovered for LED bodys
  *)
    //Fast handling for stock gridpoints
    if (bVal = LWS_NOTHING) or
        (bVal = LWS_OBJECT) or
        (bVal = LWS_CONNECTOR_HOR) or
        (bVal = LWS_CONNECTOR_VER) or
        (bVal = LWS_TEXT) or
        (bVal = LWS_LED_BODY) or
        (bVal = LWS_CONNP_TOP) or
        (bVal = LWS_CONNP_BOTTOM) or
        (bVal = LWS_VALUE) then
      ret:=LWS_P_NOTHING
    //Simple horizontal stuff
    else if (bVal = LWS_CABLE_HOR_L) or
            (bVal = LWS_CABLE_HOR_L*LWS_ON)
            then ret := LWS_P_RIGHT
    else if (bVal = LWS_CABLE_HOR_R) or
            (bVal = LWS_CABLE_HOR_R*LWS_ON)
            then ret := LWS_P_LEFT
    else if (bVal = LWS_CABLE_HOR) or
            (bVal = LWS_CABLE_HOR*LWS_ON)
            then ret := LWS_P_RIGHT or LWS_P_LEFT
    //Simple vertical stuff
    else if (bVal = LWS_CABLE_VER_T) or
            (bVal = LWS_CABLE_VER_T*LWS_ON)
            then ret := LWS_P_BOTTOM
    else if (bVal = LWS_CABLE_VER_B) or
            (bVal = LWS_CABLE_VER_B*LWS_ON)
            then ret := LWS_P_TOP
    else if (bVal = LWS_CABLE_VER) or
            (bVal = LWS_CABLE_VER*LWS_ON)
            then ret := LWS_P_TOP or LWS_P_BOTTOM
    //Edges (Vertical + horizontal)
    else if (bVal = LWS_EDGE_L_T) or
            (bVal = LWS_EDGE_L_T*LWS_ON)
            then ret := LWS_P_BOTTOM or LWS_P_RIGHT
    else if (bVal = LWS_EDGE_R_T) or
            (bVal = LWS_EDGE_R_T*LWS_ON)
            then ret := LWS_P_BOTTOM or LWS_P_LEFT
    else if (bVal = LWS_EDGE_L_B) or
            (bVal = LWS_EDGE_L_B*LWS_ON)
            then ret := LWS_P_TOP or LWS_P_RIGHT
    else if (bVal = LWS_EDGE_R_B) or
            (bVal = LWS_EDGE_R_B*LWS_ON)
            then ret := LWS_P_TOP or LWS_P_LEFT
    //Unconnected crossing
    else if (bVal = LWS_CABLE_CROSS) or
            (bVal = LWS_CABLE_CROSS*LWS_ON) or
            (bVal = LWS_CABLE_VER*LWS_ON + LWS_CABLE_HOR) or
            (bVal = LWS_CABLE_HOR*LWS_ON + LWS_CABLE_VER)
            then ret := LWS_P_TOP or LWS_P_BOTTOM or LWS_P_LEFT or LWS_P_RIGHT
    //connected crossing
    else if (bVal = LWS_CABLE_CROSS + LWS_CONNECTED) or
            (bVal = LWS_CABLE_CROSS*LWS_ON + LWS_CONNECTED)
            then ret := LWS_P_TOP or LWS_P_BOTTOM or LWS_P_LEFT or LWS_P_RIGHT or LWS_P_CONNECTED
    //partial cross
    // _|_
    else if (bVal = LWS_CABLE_VER_B + LWS_CABLE_HOR) or  //unconnected in locad
            (bVal = LWS_CABLE_VER_B*LWS_ON + LWS_CABLE_HOR) or
            (bVal = LWS_CABLE_VER_B*LWS_ON + LWS_CABLE_HOR*LWS_ON) or
            (bVal = LWS_CABLE_VER_B + LWS_CABLE_HOR*LWS_ON) or

            (bVal = LWS_CABLE_VER_B + LWS_CABLE_HOR + LWS_CONNECTED) or //connected in lc
            (bVal = LWS_ON*(LWS_CABLE_VER_B + LWS_CABLE_HOR) + LWS_CONNECTED)
            then ret := LWS_P_LEFT or LWS_P_RIGHT or LWS_P_TOP or LWS_P_CONNECTED
    //___
    // |
    else if (bVal = LWS_CABLE_VER_T + LWS_CABLE_HOR) or  //unconnected in locad
            (bVal = LWS_CABLE_VER_T*LWS_ON + LWS_CABLE_HOR) or
            (bVal = LWS_CABLE_VER_T*LWS_ON + LWS_CABLE_HOR*LWS_ON) or
            (bVal = LWS_CABLE_VER_T + LWS_CABLE_HOR*LWS_ON) or

            (bVal = LWS_CABLE_VER_T + LWS_CABLE_HOR + LWS_CONNECTED) or //connected in lc
            (bVal = LWS_ON*(LWS_CABLE_VER_T + LWS_CABLE_HOR) + LWS_CONNECTED)
            then ret := LWS_P_LEFT or LWS_P_RIGHT or LWS_P_BOTTOM or LWS_P_CONNECTED
    // |
    // |--
    // |
    else if (bVal = LWS_CABLE_HOR_L + LWS_CABLE_VER) or // unconnected in locad
            (bVal = LWS_CABLE_HOR_L*LWS_ON + LWS_CABLE_VER) or
            (bVal = LWS_CABLE_HOR_L*LWS_ON + LWS_CABLE_VER*LWS_ON) or
            (bVal = LWS_CABLE_HOR_L + LWS_CABLE_VER*LWS_ON) or

            (bVal = LWS_CABLE_HOR_L + LWS_CABLE_VER + LWS_CONNECTED) or
            (bVal = LWS_ON*(LWS_CABLE_HOR_L + LWS_CABLE_VER) + LWS_CONNECTED)
            then ret := LWS_P_TOP or LWS_P_BOTTOM or LWS_P_RIGHT or LWS_P_CONNECTED
    //   |
    // --|
    //   |
    else if (bVal = LWS_CABLE_HOR_R + LWS_CABLE_VER) or // unconnected in locad
            (bVal = LWS_CABLE_HOR_R*LWS_ON + LWS_CABLE_VER) or
            (bVal = LWS_CABLE_HOR_R*LWS_ON + LWS_CABLE_VER*LWS_ON) or
            (bVal = LWS_CABLE_HOR_R + LWS_CABLE_VER*LWS_ON) or

            (bVal = LWS_CABLE_HOR_R + LWS_CABLE_VER + LWS_CONNECTED) or
            (bVal = LWS_ON*(LWS_CABLE_HOR_R + LWS_CABLE_VER) + LWS_CONNECTED)
            then ret := LWS_P_TOP or LWS_P_BOTTOM or LWS_P_LEFT or LWS_P_CONNECTED
    //Coupled stuff
    else if (bVal = LWS_CABLE_CROSS + LWS_COUPLED) or
            (bVal = LWS_CABLE_CROSS*LWS_ON + LWS_COUPLED) or
            (bVal = LWS_CABLE_VER*LWS_ON + LWS_CABLE_HOR + LWS_COUPLED)
            then ret := LWS_P_TOP or LWS_P_BOTTOM or LWS_P_LEFT or LWS_P_RIGHT or LWS_P_COUPLED
    else if (bVal = LWS_CABLE_VER_T + LWS_CABLE_HOR + LWS_COUPLED) or
            (bVal = LWS_CABLE_VER_T*LWS_ON + LWS_CABLE_HOR + LWS_COUPLED) or
            (bVal = LWS_ON*(LWS_CABLE_VER_T + LWS_CABLE_HOR) + LWS_COUPLED)
            then ret := LWS_P_BOTTOM or LWS_P_LEFT or LWS_P_RIGHT or LWS_P_COUPLED or LWS_P_CONNECTED
    (*
    Every known case should already been detected. Anyway, it is possible that
    some cases slipped during analysis of the format. New things can be added
    here.
    *)

    else ret:=LWS_P_NOTHING;
    //Anything
    Result:=ret;
  end;
  
 //This functions provides the real loading and is called by the public
 //load functions
 function LS_LWSLoader.LoadIn(sPathIn: string):boolean;
 var  pFile:TStream;
      pComponent:^LS_LWS_Component;
      pText:^LS_LWS_Text;
      //Variables for conversion
      pPoint:^TPoint;
      pLine:^LS_Line;
      lLines,lCouple:TList;
      iProp,iTmpProp:LWS_P;
      I,X,Y,D:integer;
 begin
   //Allocate memory and initialize the lists
    New(self.lwsHead);
    self.lwsComponent := TList.Create;
    self.lwsText := TList.Create;
    New(self.lwsTail);
    pComponent := Nil;
    pText := Nil;
    //Alright, load the file

    (*
    Read the file to the according structures
    *)
    try
      pFile := TFileStream.Create(sPathIn, fmOpenRead);   //Open for reading
      try
        //Read head
        pFile.Read(self.lwsHead^,sizeof(self.lwsHead^));

        //Check if the file header belongs to a importable locad file
        if ((self.lwsHead.bColumnCount <> LWS_ROWS) and
           (self.lwsHead.bColumnCount <> $20)) or
           (self.lwsHead.bFirst <> 0) then
           //Throw a exception. The file does not seem to be a valid .lws file
           raise Exception.Create('Filetype does not seem to be .lws');
           
        //Read according to estimated number of components in head
        for I := 1 to self.lwsHead.bComponentCount do
          begin
          //Allocate space
          New(pComponent);
          //Read the component
          pFile.Read(pComponent^, sizeof(pComponent^));
          //Save to list
          self.lwsComponent.Add(pComponent);
          //Clear the pointer
          pComponent := NiL;
          end;

        //Read the tail
        pFile.Read(self.lwsTail^, sizeof(self.lwsTail^));

        //Read the text attachments
        for I := 1 to self.lwsTail.bStrings do
          begin
            //Allocate space
            New(pText);
            //Read
            pFile.Read(pText^, sizeof(pText^));
            //Save to list
            self.lwsText.Add(pText);
            //Clear the pointer
            pText := Nil;
          end;

      finally
        pFile.Free;  //In any case free the file
      end;
    except   //File couldn't be opened / read or something else happened...
      if pComponent <> Nil then Dispose(pComponent);
      if pText <> Nil then Dispose(pText);


      self.Release; //Be sure to clear everything
      Result:=False;
      exit;
    end;

    (*
    Analyse and save to pElementBox
    *)
    lLines:=TList.Create;  //Create the temporary lists we will use
    lCouple:=TList.Create;
    (*
    SEARCH THE CABLES
    *)
    for X := 1 to LWS_COLUMNS do   //Loop through columns
        for Y := 1 to LWS_ROWS do  //Lines
          begin
          iProp:=self.PropertyOf(self.lwsHead.abMesh[X,Y]);

          if iProp <> LWS_P_NOTHING then
            begin
            if (((iProp and LWS_P_RIGHT) > 0) or ((iProp and LWS_P_LEFT) > 0)) and not self.IsInLine(point(X,Y),lLines,True) then
              begin
              //We have found a new horizontal line, follow it to it's end
              New(pLine);
              //If there's a connector to the left...
              if (X>1) and ((iProp and LWS_P_LEFT) > 0) then
                pLine.pointStart.X := X-1  //the cable must have started @ x-1
              else pLine.pointStart.X := X; //else it is just x
              pLine.pointStart.Y := Y;
              D:=X; //Prepare for delta
              while D<LWS_COLUMNS do //Run through the line till the end (or break)
                begin
                //Get the properties of the observed point
                iTmpProp:=self.PropertyOf(self.lwsHead.abMesh[D,Y]);
                //Abort if there's nothing horizontal in the element
                if ((iTmpProp and LWS_P_RIGHT) = 0)  then break;
                //If "connected"
                if ((iTmpProp and LWS_P_CONNECTED) > 0) and
                   (pLine.pointStart.X <> D) then
                  begin
                    //add the cable and start a new one
                    pLine.pointEnd.X := D;  //save this position as line end
                    pLine.pointEnd.Y := Y;
                    lLines.Add(pLine);  //Add the line to the local list

                    New(pLine);
                    pLine.pointStart.X := D;
                    pLine.pointStart.Y := Y;
                  end;
                  //if "coupled"
                  if (iTmpProp and LWS_P_COUPLED) > 0 then
                    begin
                      //Add the coupling diode
                      New(pPoint);
                      pPoint.X := D;
                      pPoint.Y := Y;
                      lCouple.Add(pPoint);
                    end;

                D:=D+1; //Continue to the next thing
                end;
              pLine.pointEnd.X := D;  //save this position as line end
              pLine.pointEnd.Y := Y;
              lLines.Add(pLine);  //Add the line to the local list
              end;

            if (((iProp and LWS_P_BOTTOM) > 0) or ((iProp and LWS_P_TOP) > 0)) and not self.IsInLine(point(X,Y),lLines,False) then
             begin
              //We have found a new vertical line, follow it to it's end
              New(pLine);
              //If there's a connector to the top
              if (Y>1) and ((iProp and LWS_P_TOP) > 0) then
                pLine.pointStart.Y := Y-1
              else pLine.pointStart.Y := Y;
              pLine.pointStart.X := X;
              D:=Y; //Prepare for delta
              while D<LWS_ROWS do  //follow the column to its end (or break)
                begin
                //Get the properties of the observed point
                iTmpProp:=self.PropertyOf(self.lwsHead.abMesh[X,D]);
                //Abort if there's nothing vertical in the element
                if ((iTmpProp and LWS_P_BOTTOM) = 0)  then break;
                //If "connected"
                if ((iTmpProp and LWS_P_CONNECTED) > 0) and
                   (pLine.pointStart.Y <> D) then
                  begin
                    //add the cable and start a new one
                    pLine.pointEnd.X := X;  //save this position as line end
                    pLine.pointEnd.Y := D;
                    lLines.Add(pLine);  //Add the line to the local list

                    New(pLine);
                    pLine.pointStart.X := X;
                    pLine.pointStart.Y := D;
                  end;

                D:=D+1; //Continue to the next element
                end;
                pLine.pointEnd.X := X;  //save this position as line end
                pLine.pointEnd.Y := D;
                lLines.Add(pLine);  //Add the line to the local list
              end;
            end;
          end;
    (*
    Search for cables and relatives is completed, now add the results to the actual
    LogikSim object container.
    *)
    for pLine in lLines do
      begin
        self.mTElementBox.AddDirectLine(pLine.pointStart,pLine.pointEnd);
       end;
    //Update the list
    self.mTElementBox.RefreshLineConnectorList;
    //Cleanup
    while lLines.Count>0 do
      begin
        Dispose(lLines[0]);
        lLines.Delete(0);
      end;
    lLines.Free;

    //Add the coupling diodes
    for pPoint in lCouple do
      begin
        self.mTElementBox.SetCouplingDiode(pPoint^);
      end;
    //Cleanup
    while lCouple.Count>0 do
      begin
        Dispose(lCouple[0]);
        lCouple.Delete(0);
      end;
    lCouple.Free;
    
    (*
    Analyse the objects read from the file and directly add them into
    our LogikSim object container.
    *)
    for pComponent in self.lwsComponent do
      begin
        //In case of a simple OR
        if pComponent.bComponentType = LWS_OR then
          begin
          if pComponent.bRotation = LWS_ROT_LEFT then
            self.mTElementBox.AddOrModule(point(pComponent.rPinIn[1].wGridX+1,
                                              pComponent.rPinIn[1].wGridY+1),orLeft,3)
          else if pComponent.bRotation = LWS_ROT_RIGHT then
            self.mTElementBox.AddOrModule(point(pComponent.rPinIn[1].wGridX-1,
                                              pComponent.rPinIn[1].wGridY+1),orRight,3)
          else if pComponent.bRotation = LWS_ROT_TOP then
            self.mTElementBox.AddOrModule(point(pComponent.rPinIn[3].wGridX+1,
                                              pComponent.rPinIn[3].wGridY+1),orUp,3)
          else if pComponent.bRotation = LWS_ROT_BOTTOM then
            self.mTElementBox.AddOrModule(point(pComponent.rPinIn[3].wGridX+1,
                                              pComponent.rPinIn[3].wGridY-1),orDown,3);

          //Set negation points
          if pComponent.rPinIn[1].bNegated then
            self.mTElementBox.SetNegationPoint(point(
                                            pComponent.rPinIn[1].wGridX+1,
                                            pComponent.rPinIn[1].wGridY+1));
          if pComponent.rPinIn[2].bNegated then
            self.mTElementBox.SetNegationPoint(point(
                                            pComponent.rPinIn[2].wGridX+1,
                                            pComponent.rPinIn[2].wGridY+1));
          if pComponent.rPinIn[3].bNegated then
            self.mTElementBox.SetNegationPoint(point(
                                            pComponent.rPinIn[3].wGridX+1,
                                            pComponent.rPinIn[3].wGridY+1));

          //out
          if pComponent.rPinOut[1].bNegated then
            self.mTElementBox.SetNegationPoint(point(
                                            pComponent.rPinOut[1].wGridX+1,
                                            pComponent.rPinOut[1].wGridY+1));

          end
        //In case of a simple XOR
        else if pComponent.bComponentType = LWS_XOR then
          begin
          if pComponent.bRotation = LWS_ROT_LEFT then
            self.mTElementBox.AddXorModule(point(pComponent.rPinIn[1].wGridX+1,
                                              pComponent.rPinIn[1].wGridY+1),orLeft,3)
          else if pComponent.bRotation = LWS_ROT_RIGHT then
            self.mTElementBox.AddXorModule(point(pComponent.rPinIn[1].wGridX-1,
                                              pComponent.rPinIn[1].wGridY+1),orRight,3)
          else if pComponent.bRotation = LWS_ROT_TOP then
            self.mTElementBox.AddXorModule(point(pComponent.rPinIn[3].wGridX+1,
                                              pComponent.rPinIn[3].wGridY+1),orUp,3)
          else if pComponent.bRotation = LWS_ROT_BOTTOM then
            self.mTElementBox.AddXorModule(point(pComponent.rPinIn[3].wGridX+1,
                                              pComponent.rPinIn[3].wGridY-1),orDown,3);

         //Set negation points
          if pComponent.rPinIn[1].bNegated then
            self.mTElementBox.SetNegationPoint(point(
                                            pComponent.rPinIn[1].wGridX+1,
                                            pComponent.rPinIn[1].wGridY+1));
          if pComponent.rPinIn[2].bNegated then
            self.mTElementBox.SetNegationPoint(point(
                                            pComponent.rPinIn[2].wGridX+1,
                                            pComponent.rPinIn[2].wGridY+1));
          if pComponent.rPinIn[3].bNegated then
            self.mTElementBox.SetNegationPoint(point(
                                            pComponent.rPinIn[3].wGridX+1,
                                            pComponent.rPinIn[3].wGridY+1));

          //out
          if pComponent.rPinOut[1].bNegated then
            self.mTElementBox.SetNegationPoint(point(
                                            pComponent.rPinOut[1].wGridX+1,
                                            pComponent.rPinOut[1].wGridY+1));
          end
        //In case of a simple AND
        else if pComponent.bComponentType = LWS_AND then
          begin
          if pComponent.bRotation = LWS_ROT_LEFT then
            self.mTElementBox.AddAndModule(point(pComponent.rPinIn[1].wGridX+1,
                                              pComponent.rPinIn[1].wGridY+1),orLeft,3)
          else if pComponent.bRotation = LWS_ROT_RIGHT then
            self.mTElementBox.AddAndModule(point(pComponent.rPinIn[1].wGridX-1,
                                              pComponent.rPinIn[1].wGridY+1),orRight,3)
          else if pComponent.bRotation = LWS_ROT_TOP then
            self.mTElementBox.AddAndModule(point(pComponent.rPinIn[3].wGridX+1,
                                              pComponent.rPinIn[3].wGridY+1),orUp,3)
          else if pComponent.bRotation = LWS_ROT_BOTTOM then
            self.mTElementBox.AddAndModule(point(pComponent.rPinIn[3].wGridX+1,
                                              pComponent.rPinIn[3].wGridY-1),orDown,3);

         //Set negation points
          if pComponent.rPinIn[1].bNegated then
            self.mTElementBox.SetNegationPoint(point(
                                            pComponent.rPinIn[1].wGridX+1,
                                            pComponent.rPinIn[1].wGridY+1));
          if pComponent.rPinIn[2].bNegated then
            self.mTElementBox.SetNegationPoint(point(
                                            pComponent.rPinIn[2].wGridX+1,
                                            pComponent.rPinIn[2].wGridY+1));
          if pComponent.rPinIn[3].bNegated then
            self.mTElementBox.SetNegationPoint(point(
                                            pComponent.rPinIn[3].wGridX+1,
                                            pComponent.rPinIn[3].wGridY+1));

          //out
          if pComponent.rPinOut[1].bNegated then
            self.mTElementBox.SetNegationPoint(point(
                                            pComponent.rPinOut[1].wGridX+1,
                                            pComponent.rPinOut[1].wGridY+1));
          end
        //Full-adder
        else if pComponent.bComponentType = LWS_VA then
          begin
          if pComponent.bRotation = LWS_ROT_LEFT then
            self.mTElementBox.AddFullAdder(point(pComponent.rPinIn[1].wGridX+1,
                                              pComponent.rPinIn[1].wGridY+1),orLeft)
          //edge @ x-3,y
          else if pComponent.bRotation = LWS_ROT_RIGHT then
            self.mTElementBox.AddFullAdder(point(pComponent.rPinIn[1].wGridX-2,
                                              pComponent.rPinIn[1].wGridY+1),orRight)
          //edge @ x,y
          else if pComponent.bRotation = LWS_ROT_TOP then
            self.mTElementBox.AddFullAdder(point(pComponent.rPinIn[1].wGridX+1,
                                              pComponent.rPinIn[1].wGridY+1),orUp)
          //edge @ x,y-3
          else if pComponent.bRotation = LWS_ROT_BOTTOM then
            self.mTElementBox.AddFullAdder(point(pComponent.rPinIn[1].wGridX+1,
                                              pComponent.rPinIn[1].wGridY-2),orDown)

         //No negation possible
          end
        //Half-adder
        else if pComponent.bComponentType = LWS_HA then
          begin
          if pComponent.bRotation = LWS_ROT_LEFT then
            self.mTElementBox.AddHalfAdder(point(pComponent.rPinIn[1].wGridX+1,
                                              pComponent.rPinIn[1].wGridY+1))
          //edge @ x-3,y
          else if pComponent.bRotation = LWS_ROT_RIGHT then
            self.mTElementBox.AddHalfAdder(point(pComponent.rPinIn[1].wGridX-2,
                                              pComponent.rPinIn[1].wGridY+1),orRight)
          //edge @ x,y
          else if pComponent.bRotation = LWS_ROT_TOP then
            self.mTElementBox.AddHalfAdder(point(pComponent.rPinIn[1].wGridX+1,
                                              pComponent.rPinIn[1].wGridY+1),orUp)
          //edge @ x,y-3
          else if pComponent.bRotation = LWS_ROT_BOTTOM then
            self.mTElementBox.AddHalfAdder(point(pComponent.rPinIn[1].wGridX+1,
                                              pComponent.rPinIn[1].wGridY-2),orDown)
         //No negation possible
          end
        //Flip-Flop
        else if pComponent.bComponentType = LWS_FF then
          begin
          if pComponent.bRotation = LWS_ROT_LEFT then
            self.mTElementBox.AddFlipFlop(point(pComponent.rPinIn[1].wGridX+1,
                                              pComponent.rPinIn[1].wGridY+1),orLeft)
          //edge @ x-4,y
          else if pComponent.bRotation = LWS_ROT_RIGHT then
            self.mTElementBox.AddFlipFlop(point(pComponent.rPinIn[1].wGridX-3,
                                              pComponent.rPinIn[1].wGridY+1),orRight)
          //edge @ x,y
          else if pComponent.bRotation = LWS_ROT_TOP then
            self.mTElementBox.AddFlipFlop(point(pComponent.rPinIn[2].wGridX+1,
                                              pComponent.rPinIn[1].wGridY+1),orUp)
          //edge @ x,y-4
          else if pComponent.bRotation = LWS_ROT_BOTTOM then
            self.mTElementBox.AddFlipFlop(point(pComponent.rPinIn[2].wGridX+1,
                                              pComponent.rPinIn[1].wGridY-3),orDown);

         //Set negation points
          if pComponent.rPinIn[2].bNegated then
            self.mTElementBox.SetNegationPoint(point(
                                            pComponent.rPinIn[2].wGridX+1,
                                            pComponent.rPinIn[2].wGridY+1));
          if pComponent.rPinIn[3].bNegated then
            self.mTElementBox.SetNegationPoint(point(
                                            pComponent.rPinIn[3].wGridX+1,
                                            pComponent.rPinIn[3].wGridY+1));
          end
        //Shifting register
        else if pComponent.bComponentType = LWS_SHIFT then
          begin
          if pComponent.bRotation = LWS_ROT_LEFT then
            self.mTElementBox.AddShiftingRegister(point(pComponent.rPinIn[1].wGridX+1,
                                              pComponent.rPinIn[1].wGridY+1),orLeft)
          //edge @ x-8,y
          else if pComponent.bRotation = LWS_ROT_RIGHT then
            self.mTElementBox.AddShiftingRegister(point(pComponent.rPinIn[1].wGridX-7,
                                              pComponent.rPinIn[1].wGridY+1),orRight)
          //edge @ x,y
          else if pComponent.bRotation = LWS_ROT_TOP then
            self.mTElementBox.AddShiftingRegister(point(pComponent.rPinIn[2].wGridX+1,
                                              pComponent.rPinIn[2].wGridY+1),orUp)
          //edge @ x,y-8
          else if pComponent.bRotation = LWS_ROT_BOTTOM then
            self.mTElementBox.AddShiftingRegister(point(pComponent.rPinIn[2].wGridX+1,
                                              pComponent.rPinIn[2].wGridY-7),orDown);
          //Shifting register only clock pin allowed
          if pComponent.rPinIn[3].bNegated then
            self.mTElementBox.SetNegationPoint(point(
                                            pComponent.rPinIn[3].wGridX+1,
                                            pComponent.rPinIn[3].wGridY+1));

          end
        //Inverter element
        else if pComponent.bComponentType = LWS_INV then
          begin
          if pComponent.bRotation = LWS_ROT_LEFT then
            self.mTElementBox.AddInverter(point(pComponent.rPinIn[1].wGridX+1,
                                                pComponent.rPinIn[1].wGridY+1),orLeft)
          //edge @ x-1,y
          else if pComponent.bRotation = LWS_ROT_RIGHT then
            self.mTElementBox.AddInverter(point(pComponent.rPinIn[1].wGridX,
                                                pComponent.rPinIn[1].wGridY+1),orRight)
          //edge @ x,y
          else if pComponent.bRotation = LWS_ROT_TOP then
            self.mTElementBox.AddInverter(point(pComponent.rPinIn[1].wGridX+1,
                                                pComponent.rPinIn[1].wGridY+1),orUp)
          //edge @ x,y-1
          else if pComponent.bRotation = LWS_ROT_BOTTOM then
            self.mTElementBox.AddInverter(point(pComponent.rPinIn[1].wGridX+1,
                                                pComponent.rPinIn[1].wGridY),orDown)
          end
        //Clock generator
        else if pComponent.bComponentType = LWS_CLOCK then
          begin
            //edge @ x,y-1
            if pComponent.bRotation = LWS_ROT_RIGHT then
              self.mTElementBox.AddClockGenerator(point(
                                                pComponent.rPinOut[1].wGridX+1,
                                                pComponent.rPinOut[1].wGridY),orRight)
            //egde @ x-3,y-1
            else if pComponent.bRotation = LWS_ROT_LEFT then
              self.mTElementBox.AddClockGenerator(point(
                                                pComponent.rPinOut[1].wGridX-2,
                                                pComponent.rPinOut[1].wGridY),orLeft)
            //edge @ x-1, y-3
            else if pComponent.bRotation = LWS_ROT_TOP then
              self.mTElementBox.AddClockGenerator(point(
                                                pComponent.rPinOut[1].wGridX,
                                                pComponent.rPinOut[1].wGridY-2),orUp)
            //edge @ x-1, y
            else if pComponent.bRotation = LWS_ROT_BOTTOM then
              self.mTElementBox.AddClockGenerator(point(
                                                pComponent.rPinOut[1].wGridX,
                                                pComponent.rPinOut[1].wGridY+1),orDown);
          end
        //Decoder without clock
        else if pComponent.bComponentType = LWS_DEC then
          begin
          //edge @ x,y-3
          if pComponent.bRotation = LWS_ROT_LEFT then
            self.mTElementBox.AddDecoderNoClock(point(
                                                pComponent.rPinIn[1].wGridX+1,
                                                pComponent.rPinIn[1].wGridY-2),orLeft)
          //edge @ x-3,y-3
          else if pComponent.bRotation = LWS_ROT_RIGHT then
            self.mTElementBox.AddDecoderNoClock(point(
                                                pComponent.rPinIn[1].wGridX-2,
                                                pComponent.rPinIn[1].wGridY-2),orRight)
          //edge @ x-4,y
          else if pComponent.bRotation = LWS_ROT_TOP then
            self.mTElementBox.AddDecoderNoClock(point(
                                                pComponent.rPinIn[3].wGridX-3,
                                                pComponent.rPinIn[3].wGridY+1),orUp)
          //edge @ x-4,y-3
          else if pComponent.bRotation = LWS_ROT_BOTTOM then
            self.mTElementBox.AddDecoderNoClock(point(
                                                pComponent.rPinIn[3].wGridX-3,
                                                pComponent.rPinIn[3].wGridY-2),orDown)
          end
        //Decoder with clock
        else if pComponent.bComponentType = LWS_DEC_T then
          begin
          //edge @ x,y-3
          if pComponent.bRotation = LWS_ROT_LEFT then
            self.mTElementBox.AddDecoderClock(point(
                                                pComponent.rPinIn[1].wGridX+1,
                                                pComponent.rPinIn[1].wGridY-2),orLeft)
          //edge @ x-3,y-3
          else if pComponent.bRotation = LWS_ROT_RIGHT then
            self.mTElementBox.AddDecoderClock(point(
                                                pComponent.rPinIn[1].wGridX-2,
                                                pComponent.rPinIn[1].wGridY-2),orRight)
          //edge @ x-4,y
          else if pComponent.bRotation = LWS_ROT_TOP then
            self.mTElementBox.AddDecoderClock(point(
                                                pComponent.rPinIn[3].wGridX-3,
                                                pComponent.rPinIn[3].wGridY+1),orUp)
          //edge @ x-4,y-3
          else if pComponent.bRotation = LWS_ROT_BOTTOM then
            self.mTElementBox.AddDecoderClock(point(
                                                pComponent.rPinIn[3].wGridX-3,
                                                pComponent.rPinIn[3].wGridY-2),orDown)
          end
        //In case of a LED
        else if pComponent.bComponentType = LWS_LED then
          begin
            //Only good small-Led support up to now. Big would need some hack for
            //flawless import which surely isn't on top of my todo list :)
            if pComponent.rPinOut[8].bID = LWS_COLOR_GREEN then
              self.mTElementBox.AddLedModule(point(pComponent.rCoordinates[1].wGridX+1,
                                                pComponent.rCoordinates[1].wGridY+1),
                                            clLime)
            else if pComponent.rPinOut[8].bID = LWS_COLOR_RED then
              self.mTElementBox.AddLedModule(point(pComponent.rCoordinates[1].wGridX+1,
                                                pComponent.rCoordinates[1].wGridY+1),
                                            clRed)
            else if pComponent.rPinOut[8].bID = LWS_COLOR_YELLOW then
              self.mTElementBox.AddLedModule(point(pComponent.rCoordinates[1].wGridX+1,
                                                pComponent.rCoordinates[1].wGridY+1),
                                            $00E6FF);
          end
        //In case of a delay element
        else if pComponent.bComponentType = LWS_DELAY then
          begin
            self.mTElementBox.AddDelayModule(point(pComponent.rCoordinates[1].wGridX+1,
                                            pComponent.rCoordinates[1].wGridY+1),orLeft);
          end
        //In case of a switch
        else if pComponent.bComponentType = LWS_SWITCH then
          begin
            self.mTElementBox.AddSwitchModule(point(pComponent.rCoordinates[1].wGridX+1,
                                            pComponent.rCoordinates[1].wGridY+1));
          end;

      end;
      
      (*
      Add the Text marks
      *)
      for pText in self.lwsText do
         self.mTElementBox.AddTextModule(point(pText.bGridX+1, pText.bGridY+1), pText.sText, pText.bSize*2);

    //We finished successfully
    Result:=True;
  end;
  

  //Saves the loaded .LWS file as a .sim file named sPathOut
  function LS_LWSLoader.SaveConverted(sPathOut:string):boolean;
  var SreamData: TRakBinaryStreamData;
  begin
    //Check if we have something we could save
    if self.mTElementBox = Nil then
    begin
      Result := False;     //If not return false
      exit;
    end;

    //Save the thingy
    try
      SreamData := TRakBinaryStreamData.Create;
      try
        SreamData.ApplicationName := FileHeader;
        SreamData.ApplicationVersion := ProgramVersion;
        self.mTElementBox.SaveToBinaryTag(SreamData.AddTag('MainSimulationTab'). AddTag('ElementBox'));
        SreamData.SaveToFile(sPathOut);
        Result := true;
      finally
        SreamData.Free;
      end;
    except
      Result := false;
    end;
  end;

  //Releases TElementBox object and resets the class
  procedure LS_LWSLoader.Release;
  begin
      //If there is something and it's ours release it
      if (self.mTElementBox <> Nil) and (self.bOwnElementBox) then
        self.mTElementBox.free;

      self.mTElementBox   := Nil; //Reset reference
      self.bOwnElementBox := False;

      //Dispose the head

      if self.lwsHead<>Nil then
        begin
        Dispose(self.lwsHead);
        self.lwsHead := Nil;
        end;

      //Clear object caches

      if self.lwsComponent<>Nil then
        begin
          while self.lwsComponent.Count>0 do
            begin
              Dispose(self.lwsComponent[0]);
              self.lwsComponent.Delete(0);
            end;
          self.lwsComponent.Free;
          self.lwsComponent:=Nil;
        end;

      //Clear text cache

      if self.lwsText<>Nil then
        begin
          while self.lwsText.Count>0 do
            begin
              Dispose(self.lwsText[0]);
              self.lwsText.Delete(0);
            end;
          self.lwsText.Free;
          self.lwsText:=Nil;
        end;

      //Delete the tail
      if self.lwsTail<>NIL then
        begin
          Dispose(self.lwsTail);
          self.lwsTail := NIL;
        end;

  end;


end.
