unit RakDestroy;
{*******************************************************************************
** Copyright  Richard Kaspar Softwareentwicklung, 2000. Alle Rechte vorbehalten
** Unit         : RakDestroy
** Autor        : Richard Kaspar
** Beschreibung : RakDestroy enthlt Schnittstellen und Klassen, mit denen
                  sichergestellt werden soll, dass Objekte nur von
                  seinem Besitzer freigegeben werden knnen.
                  Beispiel:
                    TEineKlasse = class
                    private
                      FMeineInstanz: TUnterklasse;
                    public
                      ...
                      property Zugriff: TUnterklasse read FMeineInstanz;
                    end;

                    var EineKlasse: TKlasse;

                    Von aussen koennte man jetzt ueber den Aufruf
                      EineKlasse.Zugriff.Free;
                    FMeineInstanz loeschen. Sptestens wenn TEineKlasse
                    in seinem Destruktor FMeineInstanz freigeben will
                    "kracht" es.
                    Die Fehlersuche nach diesem Bug knnte sehr lange dauern.
                    Wenn man die hier zur Verfuegung gestellten Klassen verwendet,
                    dann tritt bereits beim Aufruf EineKlasse.Zugriff.Free;
                    eine Exception auf.

                    Wie muss der Code gendert werden, damit der gewnschte
                    Effekt eintritt:
                    Nehmen wir an TUnterklasse sieht folgendermasen aus:

                    TUnterklasse = class(TList)
                    ...
                    end;

                    Dann muss TUnterklasse auf folgendes umgeschrieben werden:

                    TUnterklasse = class(TList, IRakDestroyChild)
                      private
                        FDestroyChild: TRakDestroyChild;
                    ...
                      protected
                        property DestroyChild: TRakDestroyChild read FDestroyChild implements IRakDestroyChild;
                    ...
                    end;

                    constructor TUnterklasse.Create;
                    begin
                      ...
                      FDestroyChild := TRakDestroyChild.Create(self);
                      ...
                    end;

                    destructor TUnterklasse.Destroy;
                    begin
                      DestroyChild.CheckDestroy;//Diese Zeile muss die erste im Destruktor sein
                      DestroyChild.Free;
                      ...
                    end;

                    und TEineKlasse muss auf folgendes umgeschrieben werden:

                    TEineKlasse = class(TObject, IRakDestroyOwner)
                      private
                        FDestroyOwner: TRakDestroyOwner;
                        FMeineInstanz: TUnterklasse;
                    ...
                      protected
                        property DestroyOwner: TRakDestroyOwner read FDestroyOwner implements IRakDestroyOwner;
                    end;

                    constructor TEineKlasse.Create;
                    begin
                      ...
                      FDestroyOwner := TRakDestroyOwner.Create(self);
                      FMeineInstanz := TUnterklasse.Create;
                      DestroyOwner.AddChild(FMeineInstanz);
                      ...
                    end;

                    destructor TEineKlasse.Destroy;
                    begin
                      DestroyOwner.CheckDestroy;//Diese Zeile wird nur gebraucht,
                        //wenn TEineKlasse selbst eine "Unterklasse" sein kann.
                        //Dann muss diese Zeile die erste im Destruktor sein

                      // Wenn Sie von TEineKlasse noch weitere Klassen ableiten
                      // mchten dann sollten sie hier z.B. die Zeile
                      // DoDestroy;
                      // einbauen und in allen von TEineKlasse abgeleiteten
                      // Klassen nicht mehr den Destruktor Destroy sondern
                      // die Methode DoDestroy ueberschreiben.

                      DestroyOwner.Free;
                      //DestroyOwner hat FMeineInstanz bereits freigegeben.
                      //ein Aufruf von FMeineInstanz.Free an dieser Stelle
                      //wrde zu einer Excetpion fhren
                      ...
                    end;

                  In den Units RakBereich und RakReport wird dieses
                  Konzept durchgngig angewendet. In der Klasse
                  TRakChildCollection aus der Unit RakCollection
                  wird auch der Fall behandelt, dass
                  eine Klasse bereits eine Methode DoDestroy untersttzt.
** nderungen    :
  1.10.2001:   Neues Interface IRakKomposition
              Neue Klasse TRakKomposition
  6.10.2001:   Neues Interface IRakObject
              Neue Klasse  TRakInterfacedObject
  7.10.2001:   Neue Interfaces IRakOwner und IRakOwned
*******************************************************************************}
///////////////////////////////////////////////////////////////////////////////}
// Delphi 5 Implementation
//
// Copyright (c) 2000 by Richard Kaspar
// ("http://www.kasparsoft.de")
//
// Definitions:
// - "Package" refers to the collection of files distributed by
//   the Copyright Holder, and derivatives of that collection of
//   files created through textual modification.
// - "Standard Version" refers to such a Package if it has not
//   been modified, or has been modified in accordance with the
//   wishes of the Copyright Holder.
// - "Copyright Holder" is whoever is name in the copyright or
//   copyrights for the package.
// - "You" is you, if you're thinking about copying or distributing
//   this Package.
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Package"), to deal in the Package without restriction,
// including without limitation the rights to use, copy, modify,
// merge, publish, distribute, sublicense, and/or sell copies of the
// Package, and to permit persons to whom the Package is furnished
// to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Package.
//
// You may modify your copy of this Package in any way, provided
// that you insert a prominent notice in each changed file stating
// how and when you changed a file, and provided that you do at
// least one of the following:
//
// a) allow the Copyright Holder to include your modifications in
// the Standard Version of the Package.
//
// b) use the modified Package only within your corporation or
// organization.
//
// c) rename any non standard executables, units, and classes so
// the names do not conflict with standard executables, units, and
// classes, and provide a separate manual page that clearly documents
// how it differs from the standard version.
//
// d) make other distribution arrangements with the Copyright Holder.
//
// The name of the Copyright Holder may not be used to endorse or
// promote products derived from this Package without specific prior
// written permission.
//
// THE PACKAGE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// PACKAGE OR THE USE OR OTHER DEALINGS IN THE PACKAGE.
{$I RAK.INC}

interface

uses miscfunc,
  SysUtils, ContNrs, Classes;

type
  IRakDestroyOwner = interface;
  IRakOwner = interface;

{////////////////////////////////////////////////////////////////////////////////
// Interface    : IRakObject
// Beschreibung : IRakObject mchte nichts mit dem COM-Zeugs von IUnkown
                  zu tun haben. Wenn man IRakObject oder ein davon abgeleitetes
                  Interface implementieren mchte, dann sollte die Klasse
                  von TRakInterfacedObject abgeleitet sein, oder die Methoden
                  _AddRef, QueryInterface und _Release auf die gleiche
                  Weise implementieren, wie dies TRakInterfacedObject macht.
                  Dadurch wird sichergestellt, dass sich Objekte nicht
                  automatisch freigeben, wenn es keine Interface-Referenzen
                  mehr gibt. Viele Klassen, die von IRakObject abgeleitete
                  Interfaces nutzen, verlassen sich darauf.
                  Wenn jemand diese Vorgabe nicht erfllt, dann wnsche ich
                  demjeneigen jetzt schon viel Spass bein Bug-suchen :-)
// erstellt am  : 7.10.2001
////////////////////////////////////////////////////////////////////////////////
// nderungen:
    2.2.2002: Neue Methode IGetImplementationObject
////////////////////////////////////////////////////////////////////////////////}
  IRakObject = interface
    // IGetObject gibt eine Referenz auf das Object zurck und ermglicht
    // es somit wieder ber den is-operator festzustellen mit welcher
    // Klasse man es tatschlich zu tun hat. Dadurch bleiben einen wilde
    // Castings erspart, die ekelhafte Bugs erzeugen knnen (Ich kann ein Lied davon singen).
    function IGetObject: TObject;
    // IGetImplementationObject gibt eine Referenz auf das Object zurck,
    // welches das Interface implementiert. Normalerweise geben die Funktionen
    // IGetObject und IGetImplementationObject eine Referenz auf das selbe
    // Objekt zurck. Falls aber eine Klasse zwar von einen Interface abgeleitet
    // ist, aber die Implementation auf ein anderes Objekt bertrgt, dann gibt
    // IGetObject eine Referenz auf die Instanz der eigentlichen Klasse zurck und
    // IGetImplementationObject eine Refernz auf das Objekt, das die Schnittstelle implementiert.
    // Beispiel:
    // TMeinObject = class(TObject, IVonIRakObjectAbgeleitet)
    //   FImplemantationObject: TImplementiertIVonIRakObjectAbgeleitet;
    //   property ImplObj: TImplementiertIVonIRakObjectAbgeleitet read FImplemantationObject implements IVonIRakObjectAbgeleitet;
    //   ...
    // end;
    // ...
    // procedure DoSomething(IObject: IVonIRakObjectAbgeleitet);
    // var
    //   Obj: TObject;
    // begin
    //    Obj := IObject.IGetObject;
    //          //Jetzt ist Obj eine Referenz auf eine Instanz von TMeinObject
    //    Obj := IObject.IGetImplementationObject;
    //          //Jetzt ist Obj eine Referenz auf eine Instanz von TImplementiertIVonIRakObjectAbgeleitet
    function IGetImplementationObject: TObject;
  end;

{////////////////////////////////////////////////////////////////////////////////
// Interface    : IRakOwned
// Beschreibung : IRakOwned steht fr ein Objekt, das einen Besitzer hat.
                  Dieser Besitzer ist zustndig fr die Freigabe des Objektes
                  In ISetOwner muss zuerst mit die IPlugOff-Methode
                  des besitzenden Objektes aufgerufen werden, damit
                  dieses mitbekommt, dass er nicht mehr Besitzer ist.
                  Danach muss vom neuen Besitzer die IPlugOn.Methode aufgerufen
                  werden. Natrich sollte man sich auch eine Referenz auf den
                  Besitzer merken.
                  Das implmentierende Objekt muss im Destruktor den
                  Aufruf ISetOwner(nil) haben, damit das besitzende Objekt
                  keine ungltige Referenz auf das Objekt hat. Dies verhindert
                  einen Schutzverletzung, falls das Objekt nicht von
                  seinem Besitzer, sondern von irgendeinem anderem Objekt
                  freigegeben wird
// erstellt am  : 7.10.2001
////////////////////////////////////////////////////////////////////////////////
// nderungen:
////////////////////////////////////////////////////////////////////////////////}
  IRakOwned = interface(IRakObject)
    function IGetOwner: IRakOwner;
    procedure ISetOwner(Value: IRakOwner);
  end;

{////////////////////////////////////////////////////////////////////////////////
// Interface    : IRakOwner
// Beschreibung : IRakOwner steht fr ein Objekt, das beliebig viele andere
                  Objekte besitzen kann. Es ist zustndig fr die Freigabe
                  dieser Objekte.
                  In IPlugOn muss sich die Referenz auf das neue Objekt gemerkt werden
                  In IPlugOff muss die Referenz auf das getrennte Objekt aufgerumt werden
                  In keinen von diesen beiden Methoden darf dem Owned-Objekt
                  mitgeteilt werden, dass die Verbindung zwischen ihnen abgerissen
                  ist, da diese Methoden ja vom Owned-Objekt aufgerufen werden
                  und dieses somit schon wei, dass die Verbindung getrennt wurde.
// erstellt am  : 7.10.2001
////////////////////////////////////////////////////////////////////////////////
// nderungen:
////////////////////////////////////////////////////////////////////////////////}
  IRakOwner = interface(IRakObject)
  //protected
    procedure IPlugOn(NewItem: IRakOwned);
    procedure IPlugOff(DelItem: IRakOwned);
  //public
    function GetCount: integer;
    function IGetItem(Idx: integer): IRakOwned;
  end;

{////////////////////////////////////////////////////////////////////////////////
// Interface    : IRakDestroyChild
// Beschreibung : IRakDestroyChild steht fr ein Objekt, das nur von seinem
                  Besitzer freigegeben werden kann. Der Besitzer muss
                  IRakDestroyOwner implementiert haben.
// erstellt am  : ?.?.2000
////////////////////////////////////////////////////////////////////////////////
// nderungen:
////////////////////////////////////////////////////////////////////////////////}
  IRakDestroyChild = interface
    procedure CheckDestroy; stdcall;
    procedure SetOwner(AOwnerObj: TObject; AOwner: IRakDestroyOwner); stdcall;
  end;

{////////////////////////////////////////////////////////////////////////////////
// Interface    : IRakDestroyOwner
// Beschreibung : IRakDestroyOwner steht fr ein Objekt, das Besitzer von
                  Objekten ist, die IRakDestroyChild implementiert haben
// erstellt am  : ?.?.2000
////////////////////////////////////////////////////////////////////////////////
// nderungen:
////////////////////////////////////////////////////////////////////////////////}
  IRakDestroyOwner = interface(IRakDestroyChild)
    function CanDestroy(AChild: IRakDestroyChild): boolean; stdcall;
  end;

  {////////////////////////////////////////////////////////////////////////////////
  // Klasse       : TRakInterfacedObject
  // Beschreibung : TRakInterfacedObject stellt wie TInterfacedObject eine
                    Basisklasse fr die Implementation von Schnittstellen dar.
                    Im Gegensatz zu TInterfacedObject wird aber nicht COM
                    untersttzt. Es findet keine Referenzzhlung statt und
                    somit gibt sich das Objekt auch nicht automatisch frei,
                    wenn die letzte Referenz erlischt.
                    Weiterhin wird die Methode IGetObject vom Interface
                    IRakObject implementiert, womit man eine Referenz auf
                    das Objekt erhalten kann, ohne den undurchsichtigen Effekten
                    beim Casten von einem Interface zu einem Objekt ausgesetzt
                    zu sein.
                    Die von IUnknown implementierten Methoden geben keine
                    Werte mit Informationsgehalt zurck, man kann sie also
                    komplett ignorieren.
  // erstellt am  : 6.10.2001
  ////////////////////////////////////////////////////////////////////////////////
  // nderungen:
    2.2.2002: Neue Methode IGetImplementationObject
  ////////////////////////////////////////////////////////////////////////////////}
  TRakInterfacedObject = class(TObject, IRakObject)
  private
    FIObject: TObject;
    procedure ISetObject(const Value: TObject);
  public
    function _AddRef: Integer; stdcall;
    function QueryInterface(const IID: TGUID; out Obj): HResult; stdcall;
    function _Release: Integer; stdcall;

    //IGetObject gibt FIObject oder self, wenn FIObject = nil, zurck
    function IGetObject: TObject; virtual;
    //IGetImplementationObject gibt nur self zurck
    function IGetImplementationObject: TObject;

    property IObject: TObject read IGetObject write ISetObject;
  end;

  {////////////////////////////////////////////////////////////////////////////////
  // Klasse       : TRakDestroyChild
  // Beschreibung : TRakDestroyChild kann als Basisklasse fr Klassen verwendet
                    werden, die das Interface IRakDestroyChild implementieren
                    mssen.
  // erstellt am  : ?.?.2000
  ////////////////////////////////////////////////////////////////////////////////
  // nderungen:
  ////////////////////////////////////////////////////////////////////////////////}
  TRakDestroyChild = class(TObject, IRakDestroyChild)
  private
    FChild: IRakDestroyChild;// Das ist das Objekt selbst; wird im Konstruktur bergeben
    FChildObj: TObject;
    FOwner: IRakDestroyOwner;
    FOwnerObj: TObject;
  public
    constructor Create(TheChildObj: TObject; TheChild: IRakDestroyChild);

    function _AddRef: Integer; stdcall;
    function QueryInterface(const IID: TGUID; out Obj): HResult; stdcall;
    function _Release: Integer; stdcall;

    procedure CheckDestroy; stdcall;
    procedure SetOwner(AOwnerObj: TObject; AOwner: IRakDestroyOwner); stdcall;
    procedure TheChildIsDestroyed;
  end;

  {////////////////////////////////////////////////////////////////////////////////
  // Klasse       : TRakDestroyOwner
  // Beschreibung : TRakDestroyOwner kann als Basisklasse fr Klassen verwendet
                    werden, welche die Interfaces IRakDestroyChild und
                    IRakDestroyOwner implementieren mssen.
  // erstellt am  : ?.?.2000
  ////////////////////////////////////////////////////////////////////////////////
  // nderungen:
  ////////////////////////////////////////////////////////////////////////////////}
  TRakDestroyOwner = class(TRakDestroyChild, IRakDestroyOwner)
  protected
    FCanDestroy: boolean;
    FChilds: TObjectList;
  public
    constructor Create(TheChildObj: TObject; TheOwner: IRakDestroyOwner);
    destructor Destroy; override;

    procedure AddChild(AChildObj: TObject; AChild: IRakDestroyChild); virtual;
    function CanDestroy(AChild: IRakDestroyChild): boolean; virtual; stdcall;
    procedure FreeChild(AChild: IRakDestroyChild); virtual;
  end;

  {////////////////////////////////////////////////////////////////////////////////
  // Klasse       : TRakDestroyOwnerBase
  // Beschreibung : TRakDestroyOwnerBase kann als Basisklasse fr Klassen verwendet
                    werden, die das Interface IRakDestroyOwner implementieren
                    mssen.
  // erstellt am  : ?.?.2000
  ////////////////////////////////////////////////////////////////////////////////
  // nderungen:
  ////////////////////////////////////////////////////////////////////////////////}
  TRakDestroyOwnerBase = class(TObject, IRakDestroyOwner)
  private
    FDestroyOwner: TRakDestroyOwner;
    function GetDestroyOwner: TRakDestroyOwner;
  protected
    procedure DoDestroy; virtual;
    property DestroyOwner: TRakDestroyOwner read GetDestroyOwner implements IRakDestroyOwner;
  public
    constructor Create;
    destructor Destroy; override;
  end;

  ERakDoDestroyException = class(Exception)
  public
    constructor Create(Child, Owner: TObject); overload;
  end;

  ERakDestroySetOwnerException = class(Exception)
  public
    constructor Create(Child, Owner: TObject); overload;
  end;

  {////////////////////////////////////////////////////////////////////////////////
  // Interface    : IRakKomposition
  // Beschreibung : Mit IRakKomposition kann man mehrere Objekte so fest
                    aneinander binden, dass alle freigegeben werden, sobald
                    eines dieser Objekte freigegeben wird.
                    Dem Konstruktor gibt man das erste Mitglied mit.
                    Weitere Mitglieder kann man mit AddMember hinzufgen.
                    Man sollte nie eine Referenz auf TRakKomposition halten,
                    sondern immer nur Referenzen auf IRakKomposition, denn
                    sobald die letzte Referenz auf IRakKomposition weg ist,
                    wird die Instanz von TRakKomposition automatisch gelscht.
                    WICHTIG: Jedes Mitglied, muss die Methode
                            MemberDestroying(self) in seinem Destruktor aufrufen,
                            damit die Komposition mitbekommt, wenn sich ein
                            Objekt zerstrt.
                            Der erste Aufruf im Destruktor sollte allerdings
                            DeleteMember(self) sein, damit das Risiko, dass
                            eine Instanz durch Seiteneffekte zweimal freigegeben
                            wird, verringert wird.
  // erstellt am  : 20.11.2001
  ////////////////////////////////////////////////////////////////////////////////
  // nderungen:
  ////////////////////////////////////////////////////////////////////////////////}
  IRakKomposition = interface
    // MemberDestroying(self) muss von jedem Mitglied in dessen Destruktor aufgerufen werden.
    // MemberDestroying sorgt dafr, dass alle Mitglieder und die Komposition selbst freigegeben werden,
    // egal ob ber den Parameter Member ein tatschliches Mitglied bergeben wird oder nicht!
    procedure MemberDestroying(Member: TObject);

    // AddMember fgt der Komposition ein weiteres Mitglied hinzu. Nil und doppelte Mitgliedschaft wird ignoriert
    // ACHTUNG: Es drfen nur Mitglieder hinzugefgt werden, die MemberDestroying in ihrem Destruktor aufrufen,
    // sonst kann es passieren, dass Objekte zweimal freigegeben werden und es zu einer Schutzverletzung kommt
    procedure AddMember(NewMember: TObject);
    // DeleteMember nimmt ein Mitglied aus der Komposition heraus, ohne es freizugeben
    // DeleteMember(self) sollte bei jedem Mitglied die erste Anweisung im Destruktor sein.
    procedure DeleteMember(DelMember: TObject);
    // FreeMember nimmt ein Mitglied aus der Komposition heraus, gibt es frei und setzt die Referenz auf nil
    // Freigegeben wird es allerdings nur, wenn es berhaupt noch ein Mitglied der Komposition war
    procedure FreeMember(var DelMember);
    // IsMember stellt fest, ob TestMember ein Mitglied der Komposition ist
    function IsMember(TestMember: TObject): boolean;
    // IsDestroying gibt true zurck, sobald das erste mal die Methode MemberDestroying aufgerufen wurde
    function IsDestroying: boolean;
  end;

  TRakDestroyDirection = (ddFIFO, ddLIFO);

  {////////////////////////////////////////////////////////////////////////////////
  // Klasse       : TRakKomposition
  // Beschreibung : TRakKomposition implementiert IRakKomposition
  // erstellt am  : 1.10.2001
  ////////////////////////////////////////////////////////////////////////////////
  // nderungen:
                  3.11.2001: Bugfix: Beim Aufrumen der Objekte hat es vorkommen
                                      knnen, dass auf Objekte zugegriffen
                                      wurde, die bereits aufgerumt waren
                            Neu: Abgeleitete Klassen knnen jetzt Einfluss darauf
                                      nehmen, in welcher Reihenfolge die
                                      Objekte zerstrt werden.
                  4.11.2001:  nderung: MemberDestroying ist jetzt eine Funktion
                              Neu: Methoden FreeMember und IsMember
                  20.11.2001: nderung: MemberDestroying ist wieder eine Prozedur
                              interface IRakKomposition eingefhrt
  ////////////////////////////////////////////////////////////////////////////////}
  TRakKomposition = class(TInterfacedObject, IRakKomposition)
  private
    FMembers: TObjectList;
    FIsDestroying: boolean;
  protected
    function GetDestroyDirection: TRakDestroyDirection; virtual;
  public
    constructor Create(FirstMember: TObject);
    destructor Destroy; override;
  public // IRakKomposition
    procedure MemberDestroying(Member: TObject); virtual;
    procedure AddMember(NewMember: TObject); virtual;
    procedure DeleteMember(DelMember: TObject); virtual;
    procedure FreeMember(var DelMember); virtual;
    function IsMember(TestMember: TObject): boolean; virtual;
    function IsDestroying: boolean;
  end;

  {////////////////////////////////////////////////////////////////////////////////
  // Klasse       : TRakOwned
  // Beschreibung : TRakOwned implementiert IRakOwned und kann als Basisklasse
                    fr Klassen verwendeted werden, die IRakOwned implementieren
                    mssen
  // erstellt am  : 7.10.2001
  ////////////////////////////////////////////////////////////////////////////////
  // nderungen:
  ////////////////////////////////////////////////////////////////////////////////}
  TRakOwned = class(TRakInterfacedObject, IRakOwned)
  private
    FIOwner: IRakOwner;
  protected
    procedure PlugOnOwner; virtual;
    procedure PlugOffOwner; virtual;
  public //IRakOwned
    function IGetOwner: IRakOwner;
    procedure ISetOwner(Value: IRakOwner);
  public
    destructor Destroy; override;
  end;

  {////////////////////////////////////////////////////////////////////////////////
  // Klasse       : TRakOwner
  // Beschreibung : TRakOwner implementiert IRakOwner und kann als Basisklasse
                    fr Klassen verwendeted werden, die IRakOwner implementieren
                    mssen
  // erstellt am  : 7.10.2001
  ////////////////////////////////////////////////////////////////////////////////
  // nderungen:
  ////////////////////////////////////////////////////////////////////////////////}
  TRakOwner = class(TRakInterfacedObject, IRakOwner)
  private
    FList: TInterfaceList;
  protected
    procedure IPlugOn(NewItem: IRakOwned); virtual;
    procedure IPlugOff(DelItem: IRakOwned); virtual;
  public // IRakOwner
    function GetCount: integer;
    function IGetItem(Idx: integer): IRakOwned;
  public
    constructor Create;
    destructor Destroy; override;

    procedure Clear; virtual;

    property Count: integer read GetCount;
  end;

  {////////////////////////////////////////////////////////////////////////////////
  // Klasse       : TRakOwner
  // Beschreibung : TRakOwner implementiert IRakOwned und IRakOwner und kann als
                    Basisklasse fr Klassen verwendeted werden, welche die beiden
                    Schnittstellen implementieren mssen
  // erstellt am  : 7.10.2001
  ////////////////////////////////////////////////////////////////////////////////
  // nderungen:
  ////////////////////////////////////////////////////////////////////////////////}
  TRakOwnedOwner = class(TRakOwned, IRakOwner)
  private
    FList: TInterfaceList;
  protected
    procedure IPlugOn(NewItem: IRakOwned); virtual;
    procedure IPlugOff(DelItem: IRakOwned); virtual;
  public // IRakOwner
    function GetCount: integer;
    function IGetItem(Idx: integer): IRakOwned;
  public
    constructor Create;
    destructor Destroy; override;

    procedure Clear; virtual;

    property Count: integer read GetCount;
  end;

  {////////////////////////////////////////////////////////////////////////////////
  // Klasse       : TRakOwnedSingleOwner
  // Beschreibung : TRakOwnedSingleOwner implementiert IRakOwned und IRakOwner,
                    wobei er selbst nur ein IRakOwned besitzen kann.
  // erstellt am  : 7.10.2001
  ////////////////////////////////////////////////////////////////////////////////
  // nderungen:
  ////////////////////////////////////////////////////////////////////////////////}
  TRakOwnedSingleOwner = class(TRakOwned, IRakOwner)
  private
    FOwned: IRakOwned;
    procedure SetIOwned(const Value: IRakOwned);
  protected
    procedure IPlugOn(NewItem: IRakOwned); virtual;
    procedure IPlugOff(DelItem: IRakOwned); virtual;
  public // IRakOwner
    function GetCount: integer;
    function IGetItem(Idx: integer): IRakOwned;

    property IOwned: IRakOwned read FOwned write SetIOwned;
  public
    destructor Destroy; override;
  end;

  IRakLocalObject = interface (IUnknown)
    // Stop muss aufgerufen werden, nachdem das Objekt das letzte mal bentigt wurde.
    // Dies soll nur sicher stellen, dass die Compileroptimierung das Objekt nicht zu frh
    // freigibt, deshalb muss Stop auch so implementiert werden, dass die
    // Compileroptimierung diesen Funktionausfruf nicht entfernt
    procedure Stop;
    function GetObject: TObject;
  end;

  TRakLocalObject = class (TRakInterfacedObject, IRakLocalObject)
  private
    FDestroying: boolean;
    FOwner: TObject;
  public
    constructor Create(Owner: TObject);
    destructor Destroy; override;

    procedure FreeNow;
    procedure Stop;
    function GetObject: TObject;
  end;

implementation

{ TRakDestroyChild }

procedure TRakDestroyChild.CheckDestroy;
begin
  if FOwner <> nil then
  begin
    if not FOwner.CanDestroy(self) then
      raise ERakDoDestroyException.Create(FChildObj, FOwnerObj);
  end;
end;

constructor TRakDestroyChild.Create(TheChildObj: TObject; TheChild: IRakDestroyChild);
begin
  inherited Create;

  FChild := TheChild;
end;

function TRakDestroyChild.QueryInterface(const IID: TGUID;
  out Obj): HResult;
begin
  Result :=0;
end;

procedure TRakDestroyChild.SetOwner(AOwnerObj: TObject; AOwner: IRakDestroyOwner);
begin
  if FOwnerObj = nil then
  begin
    FOwner := AOwner;
    FOwnerObj := AOwnerObj;
  end
  else
    raise ERakDestroySetOwnerException.Create(FChildObj, FOwnerObj);
end;

procedure TRakDestroyChild.TheChildIsDestroyed;
begin
  FChild := nil;
end;

function TRakDestroyChild._AddRef: Integer;
begin
  Result := 1;
end;

function TRakDestroyChild._Release: Integer;
begin
  Result := 1;
end;

{ ERakDestroySetOwnerException }

constructor ERakDestroySetOwnerException.Create(Child, Owner: TObject);
begin
  inherited Create('Die Instanz von "' + Child.ClassName
    + '" hat bereits einen Besitzer, eine Instanz von "' + TObject(Owner).ClassName
    + '"!');
end;

{ ERakDestroyException }

constructor ERakDoDestroyException.Create(Child, Owner: TObject);
begin
  inherited Create('Die Instanz von "' + Child.ClassName
    + '" darf nur von seinem Besitzer, eine Instanz von "' + TObject(Owner).ClassName
    + '" geloescht werden!');
end;

{ TRakDestroyOwner }

procedure TRakDestroyOwner.AddChild(AChildObj: TObject;
  AChild: IRakDestroyChild);
begin
  AChild.SetOwner(FChildObj, IRakDestroyOwner(FChild));
  FChilds.Add(AChildObj);
end;

function TRakDestroyOwner.CanDestroy(AChild: IRakDestroyChild): boolean;
begin
  Result := FCanDestroy;
end;

constructor TRakDestroyOwner.Create(TheChildObj: TObject; TheOwner: IRakDestroyOwner);
begin
  inherited Create(TheChildObj, TheOwner);

  FCanDestroy := false;
  FChilds := TObjectList.Create;
  FChilds.OwnsObjects := true;
end;

destructor TRakDestroyOwner.Destroy;
begin
  FCanDestroy := true;
  FChilds.Free;

  inherited;
end;

procedure TRakDestroyOwner.FreeChild(AChild: IRakDestroyChild);
var
  Idx: integer;
begin
  Idx := FChilds.IndexOf(TObject(AChild));
  if Idx >= 0 then
  begin
    FCanDestroy := true;
    try
      FChilds.Delete(Idx);
    finally
      FCanDestroy := false;
    end;
  end;
end;

{ TRakDestroyOwnerBase }

constructor TRakDestroyOwnerBase.Create;
begin
  inherited;

  FDestroyOwner := TRakDestroyOwner.Create(self, self);
end;

destructor TRakDestroyOwnerBase.Destroy;
begin
  DestroyOwner.CheckDestroy;
  DoDestroy;
  DestroyOwner.Free;
  inherited;
end;

procedure TRakDestroyOwnerBase.DoDestroy;
begin
end;

function TRakDestroyOwnerBase.GetDestroyOwner: TRakDestroyOwner;
begin
{  if FDestroyOwner = nil then
    FDestroyOwner := TRakDestroyOwner.Create(self);}

  Result := FDestroyOwner;
end;


{ TRakKomposition }

procedure TRakKomposition.AddMember(NewMember: TObject);
begin
  if (NewMember <> nil) and (FMembers.IndexOf(NewMember) = -1) then
    FMembers.Add(NewMember);
end;

constructor TRakKomposition.Create(FirstMember: TObject);
begin
  inherited Create;
  FMembers := TObjectList.Create;
  FMembers.OwnsObjects := false;
  AddMember(FirstMember);
end;

procedure TRakKomposition.DeleteMember(DelMember: TObject);
begin
  FMembers.Remove(DelMember);
end;

destructor TRakKomposition.Destroy;
begin
  MemberDestroying(nil);
  FreeTObject(FMembers);
  inherited;
end;

procedure TRakKomposition.FreeMember(var DelMember);
var
  Obj: TObject;
begin
  Obj := TObject(DelMember);
  if FMembers.IndexOf(Obj) >= 0 then
  begin
    FMembers.Remove(Obj);
    Obj.Free
  end;
  TObject(DelMember) := nil;
end;

function TRakKomposition.GetDestroyDirection: TRakDestroyDirection;
begin
  Result := ddFIFO;
end;

function TRakKomposition.IsDestroying: boolean;
begin
  Result := FIsDestroying;
end;

function TRakKomposition.IsMember(TestMember: TObject): boolean;
begin
  Result := FMembers.IndexOf(TestMember) >= 0;
end;

procedure TRakKomposition.MemberDestroying(Member: TObject);
var
  Obj: TObject;
  Idx: integer;
begin
  FIsDestroying := true;

  if FMembers <> nil then
    FMembers.Remove(Member);

  while FMembers.Count > 0 do
  begin
    if GetDestroyDirection = ddFIFO then
      Idx := 0
    else
      Idx := FMembers.Count - 1;

    Obj := FMembers[Idx];
    FMembers.Delete(Idx);
    Obj.Free;
  end;
end;

{ TRakInterfacedObject }

function TRakInterfacedObject._AddRef: Integer;
begin
  Result := 1;
end;

function TRakInterfacedObject._Release: Integer;
begin
  Result := 1;
end;

function TRakInterfacedObject.IGetObject: TObject;
begin
  if FIObject = nil then
    Result := self
  else
    Result := FIObject;
end;

function TRakInterfacedObject.QueryInterface(const IID: TGUID;
  out Obj): HResult;
begin
  Result := 0;
end;

procedure TRakInterfacedObject.ISetObject(const Value: TObject);
begin
  FIObject := Value; 
end;

function TRakInterfacedObject.IGetImplementationObject: TObject;
begin
  Result := self;
end;

{ TRakOwned }

destructor TRakOwned.Destroy;
begin
  ISetOwner(nil);

  inherited;
end;

function TRakOwned.IGetOwner: IRakOwner;
begin
  Result := FIOwner;
end;

procedure TRakOwned.ISetOwner(Value: IRakOwner);
begin
  if Value <> FIOwner then
  begin
    PlugOffOwner;

    FIOwner := Value;

    PlugOnOwner;
  end;
end;

procedure TRakOwned.PlugOffOwner;
begin
  if FIOwner <> nil then
    FIOwner.IPlugOff(self);
end;

procedure TRakOwned.PlugOnOwner;
begin
  if FIOwner <> nil then
    FIOwner.IPlugOn(self);
end;

{ TRakOwner }

procedure TRakOwner.Clear;
var
  Obj: TObject;
begin
  while Count > 0 do
  begin
    Obj := IGetItem(Count-1).IGetObject;
    Obj.Free;
  end;
end;

constructor TRakOwner.Create;
begin
  inherited;

  FList := TInterfaceList.Create;
end;

destructor TRakOwner.Destroy;
begin
  Clear;

  FList.Free;
  FList := nil;

  inherited;
end;

function TRakOwner.GetCount: integer;
begin
  Result := FList.Count;
end;

function TRakOwner.IGetItem(Idx: integer): IRakOwned;
begin
  Result := IRakOwned(FList[Idx]);
end;

procedure TRakOwner.IPlugOff(DelItem: IRakOwned);
begin
  FList.Remove(DelItem);
end;

procedure TRakOwner.IPlugOn(NewItem: IRakOwned);
begin
  if (NewItem <> nil) and (FList.IndexOf(NewItem) = -1) then
    FList.Add(NewItem);
end;

{ TRakOwnedOwner }

procedure TRakOwnedOwner.Clear;
var
  Obj: TObject;
begin
  while Count > 0 do
  begin
    Obj := IGetItem(Count-1).IGetObject;
    Obj.Free;
  end;
end;

constructor TRakOwnedOwner.Create;
begin
  inherited;

  FList := TInterfaceList.Create;
end;

destructor TRakOwnedOwner.Destroy;
begin
  Clear;

  FList.Free;
  FList := nil;

  inherited;
end;

function TRakOwnedOwner.GetCount: integer;
begin
  Result := FList.Count;
end;

function TRakOwnedOwner.IGetItem(Idx: integer): IRakOwned;
begin
  Result := IRakOwned(FList[Idx]);
end;

procedure TRakOwnedOwner.IPlugOff(DelItem: IRakOwned);
begin
  FList.Remove(DelItem);
end;

procedure TRakOwnedOwner.IPlugOn(NewItem: IRakOwned);
begin
  if (NewItem <> nil) and (FList.IndexOf(NewItem) = -1) then
    FList.Add(NewItem);
end;

{ TRakOwnedSingleOwner }

destructor TRakOwnedSingleOwner.Destroy;
begin
  if FOwned <> nil then
    FOwned.IGetObject.Free;

  inherited;
end;

function TRakOwnedSingleOwner.GetCount: integer;
begin
  if FOwned = nil then
    Result := 0
  else
    Result := 1;
end;

function TRakOwnedSingleOwner.IGetItem(Idx: integer): IRakOwned;
begin
  Result := FOwned;
end;

procedure TRakOwnedSingleOwner.IPlugOff(DelItem: IRakOwned);
begin
  FOwned := nil;
end;

procedure TRakOwnedSingleOwner.IPlugOn(NewItem: IRakOwned);
begin
  FOwned := NewItem;
end;

procedure TRakOwnedSingleOwner.SetIOwned(const Value: IRakOwned);
begin
  if Value <> nil then
    Value.ISetOwner(self);
end;

{ TRakLocalObject }

constructor TRakLocalObject.Create(Owner: TObject);
begin
  inherited Create;

  FOwner := Owner;
end;

destructor TRakLocalObject.Destroy;
begin
  FDestroying := true;
  FOwner.Free;

  inherited;
end;

procedure TRakLocalObject.FreeNow;
begin
  if self <> nil then
  begin
    FOwner := nil;
    if not FDestroying then
      Free;
  end;
end;

function TRakLocalObject.GetObject: TObject;
begin
  Result := FOwner;
end;

procedure TRakLocalObject.Stop;
begin
  FDestroying := false;
end;


end.
